/*
 * HFS/2 - A Hierarchical File System Driver for OS/2
 * Copyright (C) 1996, 1997 Marcus Better
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* stubs.c
   Stubs for the FS_xxx calls
*/

#define INCL_DOSERRORS
#define INCL_NOPMAPI
#include <os2.h>

#include <dhcalls.h>

#include <string.h>
#include <dos.h>

#include "fsd.h"
#include "..\r0r3shar.h"
#include "fsh.h"
#include "r0.h"
#ifdef DEBUG
#include "log.h"
#endif

/******************************************************************************
** Name of IFS
******************************************************************************/

char pascal FS_NAME[] = "HFS";

/******************************************************************************
** Attributes of IFS
******************************************************************************/

/*
#define SUP_REMOTE
#define SUP_UNC
#define SUP_LOCK
#define SUP_LVL7
#define SUP_FILEIO
#define SUP_PAGEIO
*/

unsigned long int pascal FS_ATTRIBUTE = 0
#if defined(SUP_REMOTE)
                        | FSA_REMOTE   /* If file system is remote          */
#endif
#if defined(SUP_UNC)
                        | FSA_UNC      /* If IFS supports UNC               */
#endif
#if defined(SUP_LOCK)
                        | FSA_LOCK     /* If IFS supports file locking      */
#endif
#if defined(SUP_LVL7)
                        | FSA_LVL7     /* If IFS supports QPathInfo level 7 */
#endif
                        ;

/*-----------------------------------------------------------------------------
--
-- Volume management
--
-----------------------------------------------------------------------------*/

/******************************************************************************
**
** FS_MOUNT - Examine a volume to determine if the IFS knows its format
**
** Parameters
** ----------
** unsigned short flag          indicates function to perform
**   values:
**     MOUNT_MOUNT              mount or accept the volume
**     MOUNT_VOL_REMOVED        volume has been removed
**     MOUNT_RELEASE            release all resources associated with 
**                              the volume
**     MOUNT_ACCEPT             accept the volume in preparation for formatting
** struct vpfsi far *pvpfsi     pointer to FSD independant volume parameters
** struct vpfsd far *pvpfsd     pointer to FSD dependant volume parameters
** unsigned short hVPB          volume handle
** char far *pBoot              pointer to sector 0 data
**
******************************************************************************/

short int far pascal _export _loadds 
FS_MOUNT(unsigned short flag, 
	 struct vpfsi far *pvpfsi,
	 struct vpfsd far *pvpfsd, 
	 unsigned short hVPB, 
	 char far *pBoot)
{
  unsigned short rc;
  OPDATA opdata;
  char *pData;
  unsigned short data_len;

  if(!CPAttached)
    return ERROR_VOLUME_NOT_MOUNTED;

  /* Setup parameters */
  opdata.funccode = CPFC_MOUNT;
  opdata.p.mount.flag = flag;
  opdata.p.mount.vpfsi = *pvpfsi;
  opdata.p.mount.vpfsd = *pvpfsd;
  opdata.p.mount.vol_descr = 0;
  opdata.p.mount.hVPB = hVPB;
  if( flag==MOUNT_MOUNT ) {
    pData = pBoot;
    data_len = 512;
  }
  else {
    pData = NULL;
    data_len = 0;
  }

  rc = do_request(&opdata, &pData, data_len);
  if (rc != NO_ERROR)
    return rc;

  /* Copy results */
  *pvpfsi = opdata.p.mount.vpfsi;
  *pvpfsd = opdata.p.mount.vpfsd;

  end_request(opdata.request_id);
  return opdata.rc;
}

/******************************************************************************
**
** FS_ATTACH - Attach or detach a drive or device
**
** Parameters
** ----------
** unsigned short flag                  indicates attaching/detaching
**   values:
**     FSA_ATTACH               attach drive/device
**     FSA_DETACH               detach drive/device
**     FSA_ATTACH_INFO          return info on attached drive/device
** char far *pDev                       drive or device that is being attached/detached
** struct vpfsd far *pvpfsd             pointer to FSD dependant volume parameters
** struct cdfsd far *pcdfsd             pointer to FSD dependant current directory
** void far *pParm                      UNVERIFIED pointer to FSD dependant attachment info
** unsigned short far *pLen             length of area pointed to by pParm
**
******************************************************************************/

#pragma argsused
short int far pascal _export _loadds 
FS_ATTACH(unsigned short flag,
	  char far *pDev,
	  struct vpfsd far *pvpfsd,
	  struct cdfsd far *pcdfsd,
	  void far *pParm,
	  unsigned short far *pLen) 
{
  return ERROR_NOT_SUPPORTED;
}

/******************************************************************************
**
** FS_FSINFO - Get/Set file system information
**
** Parameters
** ----------
** unsigned short flag                  indicates function to perform
**   values:
**     INFO_RETREIVE                    retrieve information
**     INFO_SET                         set information
** unsigned short hVPB                  volume handle
** char far *pData                      UNVERIFIED pointer to data buffer
** unsigned short cbData                length of data buffer
** unsigned short level                 type of information to return
**
******************************************************************************/

#pragma argsused
short int far pascal _export _loadds 
FS_FSINFO(unsigned short flag, unsigned short hVPB,
	  char far *pData, unsigned short cbData,
	  unsigned short level)
{
  unsigned short rc;
  OPDATA opdata;
  struct vpfsi *pvpfsi;
  struct vpfsd *pvpfsd;
  char *pdata;

  if (!CPAttached)
    return ERROR_NOT_READY;

  /* Verify parameters */
  if (flag == INFO_SET && cbData != 0)
    rc = FSH_PROBEBUF(PB_OPREAD, pData, cbData);
  else if (cbData != 0)
    rc = FSH_PROBEBUF(PB_OPWRITE, pData, cbData);
  if (rc != NO_ERROR)
     return rc;

  /* Setup parameters */
  opdata.funccode = CPFC_FSINFO;
  opdata.p.fsinfo.flag = flag;
  opdata.p.fsinfo.hVPB = hVPB;
  FSH_GETVOLPARM(hVPB, &pvpfsi, &pvpfsd);
  opdata.p.fsinfo.vpfsi = *pvpfsi;
  opdata.p.fsinfo.vpfsd = *pvpfsd;
  opdata.p.fsinfo.cbData = cbData;
  opdata.p.fsinfo.level = level;
  pdata = pData;

  rc = do_request(&opdata, &pdata, cbData);
  if (rc != NO_ERROR)
    return rc;

  /* Copy results */
  if (flag == INFO_RETRIEVE)
    memcpy(pData, pdata, cbData);
  else if(flag == INFO_SET)
    *pvpfsi = opdata.p.fsinfo.vpfsi;

  end_request(opdata.request_id);
  return opdata.rc;
}


/******************************************************************************
**
** FS_FLUSHBUF - Flush buffers for a specified volume
**
** Parameters
** ----------
** unsigned short hVPB         handle to volume to flush
** unsigned short flag         indicates whether to discard or retain cache
**   values:
**     FLUSH_RETAIN     retain cached information
**     FLUSH_DISCARD    discard cached information
**
******************************************************************************/

#pragma argsused
short int far pascal _export _loadds 
FS_FLUSHBUF(unsigned short hVPB, 
	    unsigned short flag)
{
  unsigned short rc;
  OPDATA opdata;
  struct vpfsi *pvpfsi;
  struct vpfsd *pvpfsd;

  if (!CPAttached)
    return ERROR_NOT_READY;

  opdata.funccode = CPFC_FLUSHBUF;
  FSH_GETVOLPARM(hVPB, &pvpfsi, &pvpfsd);
  opdata.p.flushbuf.vpfsd = *pvpfsd;
  opdata.p.flushbuf.flag = flag;

  rc = do_request(&opdata, NULL, 0);
  if (rc != NO_ERROR)
    return rc;

  *pvpfsd = opdata.p.flushbuf.vpfsd;

  end_request(opdata.request_id);
  return opdata.rc;
}



/*-----------------------------------------------------------------------------
--
-- Directory management
--
-----------------------------------------------------------------------------*/



/******************************************************************************
**
** FS_CHDIR - Change current directory
**
** Parameters
** ----------
** unsigned short flag         indicates flavor of call
**   values:
**     CD_EXPLICIT      creating a new current directory
**     CD_VERIFY        verifying a current directory
**     CD_FREE          freeing an instance of a current directory
** struct cdfsi far *pcdfsi    pointer to FSD independant current directory
** struct cdfsd far *pcdfsd    pointer to FSD dependant current directory
** char far *pDir              pointer to directory to change to
** unsigned short iCurDirEnd   offset to the end of the current 
**                             directory in pDir
**
******************************************************************************/

#pragma argsused
short int far pascal _export _loadds 
FS_CHDIR(unsigned short flag, 
	 struct cdfsi far *pcdfsi,
	 struct cdfsd far *pcdfsd, 
	 char far *pDir,
	 unsigned short iCurDirEnd)
{
  unsigned short rc;
  OPDATA opdata;
  struct vpfsi *pvpfsi;
  struct vpfsd *pvpfsd;

  if (!CPAttached)
    return ERROR_NOT_READY;

  opdata.funccode = CPFC_CHDIR;
  opdata.p.chdir.flag = flag;
  if (flag != CD_FREE) {
    opdata.p.chdir.cdfsi = *pcdfsi;
    FSH_GETVOLPARM(pcdfsi->cdi_hVPB, &pvpfsi, &pvpfsd);
    opdata.p.chdir.vpfsd = *pvpfsd;
  }
  opdata.p.chdir.cdfsd = *pcdfsd;
  if (flag == CD_EXPLICIT)
    strncpy(opdata.p.chdir.Dir, pDir, sizeof(opdata.p.chdir.Dir));
  opdata.p.chdir.iCurDirEnd = iCurDirEnd;

  rc = do_request(&opdata, NULL, 0);
  if (rc != NO_ERROR)
    return rc;

  *pcdfsd = opdata.p.chdir.cdfsd;
  if(flag!=CD_FREE)
    *pvpfsd = opdata.p.chdir.vpfsd;

  end_request(opdata.request_id);
  return opdata.rc;
}


/******************************************************************************
**
** FS_MKDIR - Make a new directory
**
** Parameters
** ----------
** struct cdfsi far *pcdfsi     pointer to FSD independant current directory
** struct cdfsd far *pcdfsd     pointer to FSD dependant current directory
** char far *pName              pointer to directory name to create
** unsigned short iCurDirEnd    offset to the end of the current directory
**                              in pName
** char *pEABuf                 UNVERIFIED pointer to EAs to attach
**                              to new directory
** unsigned short flags         0x40 = directory is non 8.3 filename
**
******************************************************************************/

#pragma argsused
short int far pascal _export _loadds 
FS_MKDIR(struct cdfsi far *pcdfsi, 
	 struct cdfsd far *pcdfsd,
	 char far *pName, 
	 unsigned short iCurDirEnd,
	 char far *pEABuf, 
	 unsigned short flags)
{
  unsigned short rc;
  OPDATA opdata;
  struct vpfsi *pvpfsi;
  struct vpfsd *pvpfsd;
  char *pData;
  unsigned short data_len;

  if (!CPAttached)
    return ERROR_NOT_READY;

  /* Verify parameters */
  if (FP_OFF(pEABuf)) {
    rc = FSH_PROBEBUF(PB_OPREAD, pEABuf, sizeof(EAOP));
    if (rc != NO_ERROR)
      return rc;

    /* Check that at least the first ULONG is accessible */
    rc = FSH_PROBEBUF(PB_OPREAD, ((EAOP *)pEABuf)->fpFEAList, sizeof(ULONG));
    if (rc != NO_ERROR)
      return rc;

    /* Check that it is <=64K */
    if (((EAOP *)pEABuf)->fpFEAList->cbList > 0xffff)
      return ERROR_INVALID_PARAMETER;

    /* Check that we can access the whole list */
    rc = FSH_PROBEBUF(PB_OPREAD, ((EAOP *)pEABuf)->fpFEAList,
                      ((EAOP *)pEABuf)->fpFEAList->cbList);
    if (rc != NO_ERROR)
      return rc;
  }

  opdata.funccode = CPFC_MKDIR;
  FSH_GETVOLPARM(pcdfsi->cdi_hVPB, &pvpfsi, &pvpfsd);
  opdata.p.mkdir.vpfsd = *pvpfsd;
  opdata.p.mkdir.cdfsi = *pcdfsi;
  opdata.p.mkdir.cdfsd = *pcdfsd;
  strncpy(opdata.p.mkdir.Name, pName, sizeof(opdata.p.mkdir.Name));
  opdata.p.mkdir.iCurDirEnd = iCurDirEnd;
  if (FP_OFF(pEABuf)) {
    pData = (char *)(((EAOP *)pEABuf)->fpFEAList);
    data_len = ((EAOP *)pEABuf)->fpFEAList->cbList;
  }
  else {
    pData = NULL;
    data_len = 0;
  }
  opdata.p.mkdir.flags = flags;

  rc = do_request(&opdata, &pData, data_len);
  if (rc != NO_ERROR)
    return rc;

  *pcdfsd = opdata.p.mkdir.cdfsd;
  *pvpfsd = opdata.p.chdir.vpfsd;
  if (FP_OFF(pEABuf))
    ((EAOP *)pEABuf)->oError = opdata.p.mkdir.oError;

  end_request(opdata.request_id);
  return opdata.rc;
}


/******************************************************************************
**
** FS_RMDIR - Delete directory
**
** Parameters
** ----------
** struct cdfsi far *pcdfsi     pointer to FSD independant current directory
** struct cdfsd far *pcdfsd     pointer to FSD dependant current directory
** char far *pName              pointer to directory name to delete
** unsigned short iCurDirEnd    offset to the end of the current directory 
**                              in pName
**
******************************************************************************/

#pragma argsused
short int far pascal _export _loadds 
FS_RMDIR(struct cdfsi far *pcdfsi, 
	 struct cdfsd far *pcdfsd,
	 char far *pName, 
	 unsigned short iCurDirEnd)
{
  unsigned short rc;
  OPDATA opdata;
  struct vpfsi *pvpfsi;
  struct vpfsd *pvpfsd;

  if (!CPAttached)
    return ERROR_NOT_READY;

  opdata.funccode = CPFC_RMDIR;
  FSH_GETVOLPARM(pcdfsi->cdi_hVPB, &pvpfsi, &pvpfsd);
  opdata.p.rmdir.vpfsd = *pvpfsd;
  opdata.p.rmdir.cdfsi = *pcdfsi;
  opdata.p.rmdir.cdfsd = *pcdfsd;
  strncpy(opdata.p.rmdir.Name, pName, sizeof(opdata.p.rmdir.Name));
  opdata.p.rmdir.iCurDirEnd = iCurDirEnd;

  rc = do_request(&opdata, NULL, 0);
  if (rc != NO_ERROR)
    return rc;

  *pcdfsd = opdata.p.rmdir.cdfsd;
  *pvpfsd = opdata.p.chdir.vpfsd;

  end_request(opdata.request_id);
  return opdata.rc;
}



/*-----------------------------------------------------------------------------
--
-- File management
--
-----------------------------------------------------------------------------*/



/******************************************************************************
**
** FS_CHGFILEPTR - Change current location in file
**
** Parameters
** ----------
** struct sffsi far *psffsi      pointer to FSD independant file instance
** struct sffsd far *psffsd      pointer to FSD dependant file instance
** long offset                   signed offset
** unsigned short type           indicates seek type
**   values:
**     CFP_RELBEGIN     move pointer relative to begining of file
**     CFP_RELCUR       move pointer relative to current position in file
**     CFP_RELEND       move pointer relative to end of file
** unsigned short IOflag         bitfield of I/O suggestions
**   values:
**     IOFL_WRITETHRU   write all updated data before returning
**     IOFL_NOCACHE     don't cache any new data
**
******************************************************************************/

#pragma argsused
short int far pascal _export _loadds 
FS_CHGFILEPTR(struct sffsi far *psffsi, 
	      struct sffsd far *psffsd,
	      long offset, 
	      unsigned short type, 
	      unsigned short IOflag)
{
  unsigned short rc;
  OPDATA opdata;
  struct vpfsi *pvpfsi;
  struct vpfsd *pvpfsd;

  if (!CPAttached)
    return ERROR_NOT_READY;

  opdata.funccode = CPFC_CHGFILEPTR;
  FSH_GETVOLPARM(psffsi->sfi_hVPB, &pvpfsi, &pvpfsd);
  opdata.p.chgfileptr.vpfsd = *pvpfsd;
  opdata.p.chgfileptr.sffsi = *psffsi;
  opdata.p.chgfileptr.sffsd = *psffsd;
  opdata.p.chgfileptr.offset = offset;
  opdata.p.chgfileptr.type = type;
  opdata.p.chgfileptr.IOflag = IOflag;

  rc = do_request(&opdata, NULL, 0);
  if (rc != NO_ERROR)
    return rc;

  /* Copy results */
  *pvpfsd = opdata.p.chgfileptr.vpfsd;    
  *psffsi = opdata.p.chgfileptr.sffsi;
  *psffsd = opdata.p.chgfileptr.sffsd;

  end_request(opdata.request_id);
  return opdata.rc;
}


/******************************************************************************
**
** FS_CLOSE - Close an open file
**
** Parameters
** ----------
** unsigned short type            indicates close type
**   values:
**     FS_CL_ORDINARY  this is not the final close of the file
**     FS_CL_FORPROC   this is the final close of the file for the process
**     FS_CL_FORSYS    this is the final close of the file for the whole system
** unsigned short IOflag          bitfield of I/O suggestions
**   values:
**     IOFL_WRITETHRU  write all updated data before returning
**     IOFL_NOCACHE    don't cache any new data
** struct sffsi far *psffsi       pointer to FSD independent file instance
** struct sffsd far *psffsd       pointer to FSD dependent file instance
**
******************************************************************************/

#pragma argsused
short int far pascal _export _loadds 
FS_CLOSE(unsigned short type, 
	 unsigned short IOflag,
	 struct sffsi far *psffsi, 
	 struct sffsd far *psffsd)
{
  unsigned short rc;
  OPDATA opdata;
  struct vpfsi *pvpfsi;
  struct vpfsd *pvpfsd;

  if (!CPAttached)
    return ERROR_NOT_READY;

  opdata.funccode = CPFC_CLOSE;
  FSH_GETVOLPARM(psffsi->sfi_hVPB, &pvpfsi, &pvpfsd);
  opdata.p.close.vpfsd = *pvpfsd;
  opdata.p.close.type = type;
  opdata.p.close.IOflag = IOflag;
  opdata.p.close.sffsi = *psffsi;
  opdata.p.close.sffsd = *psffsd;

  rc = do_request(&opdata, NULL, 0);
  if (rc != NO_ERROR)
    return rc;

  /* Copy results */
  *pvpfsd = opdata.p.close.vpfsd;
  *psffsi = opdata.p.close.sffsi;
  *psffsd = opdata.p.close.sffsd;

  end_request(opdata.request_id);
  return opdata.rc;
}


/******************************************************************************
**
** FS_COMMIT - Commit a file to disk
**
** Parameters
** ----------
** unsigned short type           indicates commit type
**   values:
**     FS_COMMIT_ONE    commit this one file
**     FS_COMMIT_ALL    commit all files
** unsigned short IOflag         bitfield of I/O suggestions
**   values:
**     IOFL_WRITETHRU   write all updated data before returning
**     IOFL_NOCACHE     don't cache any new data
** struct sffsi far *psffsi      pointer to FSD independant file instance
** struct sffsd far *psffsd      pointer to FSD dependant file instance
**
******************************************************************************/

#pragma argsused
short int far pascal _export _loadds 
FS_COMMIT(unsigned short type, 
	  unsigned short IOflag,
	  struct sffsi far *psffsi, 
	  struct sffsd far *psffsd)
{
  unsigned short rc;
  OPDATA opdata;
  struct vpfsi *pvpfsi;
  struct vpfsd *pvpfsd;

  if (!CPAttached)
    return ERROR_NOT_READY;

  opdata.funccode = CPFC_COMMIT;
  FSH_GETVOLPARM(psffsi->sfi_hVPB, &pvpfsi, &pvpfsd);
  opdata.p.commit.vpfsd = *pvpfsd;
  opdata.p.commit.type = type;
  opdata.p.commit.IOflag = IOflag;
  opdata.p.commit.sffsi = *psffsi;
  opdata.p.commit.sffsd = *psffsd;

  rc = do_request(&opdata, NULL, 0);
  if (rc != NO_ERROR)
    return rc;

  /* Copy results */
  *pvpfsd = opdata.p.commit.vpfsd;
  *psffsi = opdata.p.commit.sffsi;
  *psffsd = opdata.p.commit.sffsd;

  end_request(opdata.request_id);
  return opdata.rc;
}


/******************************************************************************
**
** FS_COPY - Copy a file
**
** Parameters
** ----------
** unsigned short flag           indicates flavor of call
**   values:
**     DCPY_EXISTING    if destination file exists, replace it
**     DCPY_APPEND      source file should be appended to destination file
** struct cdfsi far *pcdfsi      pointer to FSD independant current directory
** struct cdfsd far *pcdfsd      pointer to FSD dependant current directory
** char far *pSrc                pointer to source filename
** unsigned short iSrcCurDirEnd  offset to the end of the current directory 
**                               in pSrc
** char far *pDst                pointer to destination filename
** unsigned short iDstCurDirEnd  offset to the end of the current directory 
**                               in pDst
** unsigned short nameType       0x40 = destination is non 8.3 filename
**
******************************************************************************/

#pragma argsused
short int far pascal _export _loadds 
FS_COPY(unsigned short flag, 
	struct cdfsi far *pcdfsi,
	struct cdfsd far *pcdfsd, 
	char far *pSrc,
	unsigned short iSrcCurDirEnd, 
	char far *pDst,
	unsigned short iDstCurDirEnd, 
	unsigned short nameType)
{
  unsigned short rc;
  OPDATA opdata;
  struct vpfsi *pvpfsi;
  struct vpfsd *pvpfsd;

  if (!CPAttached)
    return ERROR_NOT_READY;

  opdata.funccode = CPFC_COPY;
  FSH_GETVOLPARM(pcdfsi->cdi_hVPB, &pvpfsi, &pvpfsd);
  opdata.p.copy.vpfsd = *pvpfsd;
  opdata.p.copy.flag = flag;
  opdata.p.copy.cdfsi = *pcdfsi;
  opdata.p.copy.cdfsd = *pcdfsd;
  strncpy(opdata.p.copy.Src, pSrc, sizeof(opdata.p.copy.Src));
  opdata.p.copy.iSrcCurDirEnd = iSrcCurDirEnd;
  strncpy(opdata.p.copy.Dst, pDst, sizeof(opdata.p.copy.Dst));
  opdata.p.copy.iDstCurDirEnd = iDstCurDirEnd;
  opdata.p.copy.nameType = nameType;

  rc = do_request(&opdata, NULL, 0);
  if (rc != NO_ERROR)
    return rc;

  *pvpfsd = opdata.p.copy.vpfsd;
  *pcdfsd = opdata.p.copy.cdfsd;

  end_request(opdata.request_id);
  return opdata.rc;
}


/******************************************************************************
**
** FS_DELETE - Delete a file
**
** Parameters
** ----------
** struct cdfsi far *pcdfsi      pointer to FSD independant current directory
** struct cdfsd far *pcdfsd      pointer to FSD dependant current directory
** char far *pFile               pointer to filename to delete
** unsigned short iCurDirEnd     offset to the end of the current directory 
**                               in pFile
**
******************************************************************************/

#pragma argsused
short int far pascal _export _loadds 
FS_DELETE(struct cdfsi far *pcdfsi, 
	  struct cdfsd far *pcdfsd,
	  char far *pFile, 
	  unsigned short iCurDirEnd)
{
  unsigned short rc;
  OPDATA opdata;
  struct vpfsi *pvpfsi;
  struct vpfsd *pvpfsd;

  if (!CPAttached)
    return ERROR_NOT_READY;

  opdata.funccode = CPFC_DELETE;
  FSH_GETVOLPARM(pcdfsi->cdi_hVPB, &pvpfsi, &pvpfsd);
  opdata.p.fsdelete.vpfsd = *pvpfsd;
  opdata.p.fsdelete.cdfsi = *pcdfsi;
  opdata.p.fsdelete.cdfsd = *pcdfsd;
  strncpy(opdata.p.fsdelete.File, pFile, 
	  sizeof(opdata.p.fsdelete.File));
  opdata.p.fsdelete.iCurDirEnd = iCurDirEnd;

  rc = do_request(&opdata, NULL, 0);
  if (rc != NO_ERROR)
    return rc;

  /* Copy results */
  *pvpfsd = opdata.p.fsdelete.vpfsd;
  *pcdfsd = opdata.p.fsdelete.cdfsd;

  end_request(opdata.request_id);
  return opdata.rc;
}


/******************************************************************************
**
** FS_FILEATTRIBUTE - Get/Set DOS file attributes
**
** Parameters
** ----------
** unsigned short flag           indicates flavor of call
**   values:
**     FA_RETRIEVE      retrieve attribute
**     FA_SET           set attribute
** struct cdfsi far *pcdfsi      pointer to FSD independant current directory
** struct cdfsd far *pcdfsd      pointer to FSD dependant current directory
** char far *pName               pointer to filename
** unsigned short iCurDirEnd     offset to the end of the current directory 
**                               in pName
** unsigned short far *pAttr     pointer to the attribute
**
******************************************************************************/

#pragma argsused
short int far pascal _export _loadds 
FS_FILEATTRIBUTE(unsigned short flag, 
		 struct cdfsi far *pcdfsi,
		 struct cdfsd far *pcdfsd, 
		 char far *pName,
		 unsigned short iCurDirEnd, 
		 unsigned short far *pAttr)
{
  unsigned short rc;
  OPDATA opdata;
  struct vpfsi *pvpfsi;
  struct vpfsd *pvpfsd;

  if (!CPAttached)
    return ERROR_NOT_READY;

  opdata.funccode = CPFC_FILEATTRIBUTE;
  FSH_GETVOLPARM(pcdfsi->cdi_hVPB, &pvpfsi, &pvpfsd);
  opdata.p.fileattribute.vpfsd = *pvpfsd;
  opdata.p.fileattribute.flag = flag;
  opdata.p.fileattribute.cdfsi = *pcdfsi;
  opdata.p.fileattribute.cdfsd = *pcdfsd;
  strncpy(opdata.p.fileattribute.Name, pName, 
	  sizeof(opdata.p.fileattribute.Name));
  opdata.p.fileattribute.iCurDirEnd = iCurDirEnd;
  opdata.p.fileattribute.Attr = *pAttr;

  rc = do_request(&opdata, NULL, 0);
  if (rc != NO_ERROR)
    return rc;

  if (flag == FA_RETRIEVE)
    *pAttr = opdata.p.fileattribute.Attr;
  *pcdfsd = opdata.p.fileattribute.cdfsd;
  *pvpfsd = opdata.p.fileattribute.vpfsd;

  end_request(opdata.request_id);
  return opdata.rc;
}


/******************************************************************************
**
** FS_FILEINFO - Get/Set file information
**
** Parameters
** ----------
** unsigned short flag           indicates flavor of call
**   values:
**     FI_RETRIEVE      retrieve information
**     FI_SET           set information
** struct sffsi far *psffsi      pointer to FSD independant file instance
** struct sffsd far *psffsd      pointer to FSD dependant file instance
** unsigned short level          level of information to get/set
** char far *pData               UNVERIFIED? pointer to information area
** unsigned short cbData         size of area pointed to by pData
** unsigned short IOflag         bitfield of I/O suggestions
**   values:
**     IOFL_WRITETHRU   write all updated data before returning
**     IOFL_NOCACHE     don't cache any new data
**
******************************************************************************/

#pragma argsused
short int far pascal _export _loadds 
FS_FILEINFO(unsigned short flag, 
	    struct sffsi far *psffsi,
	    struct sffsd far *psffsd, 
	    unsigned short level,
	    char far *pData, 
	    unsigned short cbData,
	    unsigned short IOflag)
{
  unsigned short rc;
  OPDATA opdata;
  struct vpfsi *pvpfsi;
  struct vpfsd *pvpfsd;
  char *pdata;
  unsigned short data_len;

  /* Verify parameters */
  if (flag == FI_SET && cbData != 0)
    rc = FSH_PROBEBUF(PB_OPREAD, pData, cbData);
  else if (cbData != 0)
    rc = FSH_PROBEBUF(PB_OPWRITE, pData, cbData);
  if (rc != NO_ERROR)
    return rc;

  /* Special EA check */
  if (flag == FI_RETRIEVE && (level == 3 || level == 4)) {
    /* Check to make sure the buffer's at least big enough for an EAOP */
    if (cbData < sizeof(EAOP))
      return ERROR_INVALID_PARAMETER;

    if(level==3) {
      /* Check accessability of GEA */

      /* Check that at least the first ULONG is accessible */
      rc = FSH_PROBEBUF(PB_OPREAD, ((EAOP *)pData)->fpGEAList, sizeof(ULONG));
      if (rc != NO_ERROR)
        return rc;

      /* Check that it is <=64K */
      if (((EAOP *)pData)->fpGEAList->cbList > 0xffff)
        return ERROR_INVALID_PARAMETER;

      if (((EAOP *)pData)->fpGEAList->cbList == 0)
	return ERROR_BAD_FORMAT;

      /* Check that we can access the whole list */
      rc = FSH_PROBEBUF(PB_OPREAD, ((EAOP *)pData)->fpGEAList,
                        ((EAOP *)pData)->fpGEAList->cbList);
      if (rc != NO_ERROR)
        return rc;
    }
    
    /* Same thing for FEA */

    /* Check that at least the first ULONG is accessible */
    rc = FSH_PROBEBUF(PB_OPWRITE, ((EAOP *)pData)->fpFEAList, sizeof(ULONG));
    if (rc != NO_ERROR)
      return rc;

    /* Check that it is <=64K */
    if (((EAOP *)pData)->fpFEAList->cbList > 0xffff)
      return ERROR_INVALID_PARAMETER;

    /* Check that we can access the whole list */
    rc = FSH_PROBEBUF(PB_OPWRITE, ((EAOP *)pData)->fpFEAList,
                      ((EAOP *)pData)->fpFEAList->cbList);
    if (rc != NO_ERROR)
      return rc;
  }
  
  if (flag == FI_SET && level == 2) {
    /* Check that at least the first ULONG is accessible */
    rc = FSH_PROBEBUF(PB_OPREAD, ((EAOP *)pData)->fpFEAList, sizeof(ULONG));
    if (rc != NO_ERROR)
      return rc;

    /* Check that it is <=64K */
    if (((EAOP *)pData)->fpFEAList->cbList > 0xffff)
      return ERROR_INVALID_PARAMETER;

    /* Check that we can access the whole list */
    rc = FSH_PROBEBUF(PB_OPREAD, ((EAOP *)pData)->fpFEAList,
                      ((EAOP *)pData)->fpFEAList->cbList);
    if (rc != NO_ERROR)
      return rc;
  }

  if (!CPAttached)
    return ERROR_NOT_READY;

  /* Setup parameters */
  opdata.funccode = CPFC_FILEINFO;
  FSH_GETVOLPARM(psffsi->sfi_hVPB, &pvpfsi, &pvpfsd);
  opdata.p.fileinfo.vpfsd = *pvpfsd;
  opdata.p.fileinfo.flag = flag;
  opdata.p.fileinfo.sffsi = *psffsi;
  opdata.p.fileinfo.sffsd = *psffsd;
  opdata.p.fileinfo.level = level;
  /* Copy either the buffer, FEA or GEA list as appropriate */
  if (flag == FI_RETRIEVE && level == 3) {
    pdata = (char *)(((EAOP *)pData)->fpGEAList);
    data_len = ((EAOP *)pData)->fpGEAList->cbList;
    /* cbData is FEA list size, as FEA list will be returned in buffer */
    opdata.p.fileinfo.cbData = ((EAOP *)pData)->fpFEAList->cbList;
  } 
  else if(level==4) {
    opdata.p.fileinfo.cbData = ((EAOP *)pData)->fpFEAList->cbList;
    pdata = NULL;
    data_len = 0;
  }
  else if (flag == FI_SET && level == 2) {
    pdata = (char *)(((EAOP *)pData)->fpFEAList);
    data_len = ((EAOP *)pData)->fpFEAList->cbList;
    opdata.p.fileinfo.cbData = ((EAOP *)pData)->fpFEAList->cbList;
  } 
  else {
    pdata = pData;
    data_len = cbData;
    opdata.p.fileinfo.cbData = cbData;
  }

  opdata.p.fileinfo.IOflag = IOflag;
  
  rc = do_request(&opdata, &pdata, data_len);
  if (rc != NO_ERROR)
    return rc;

  /* Copy results */
  if (flag == FI_RETRIEVE)
    if (level==3 || level==4)
      memcpy(((EAOP *)pData)->fpFEAList, pdata,
	     ((FEALIST *)pData)->cbList);
    else
      memcpy(pData, pdata, cbData);
  
  if (flag == FI_SET && level == 2)
    ((EAOP *)pData)->oError = opdata.p.fileinfo.oError;

  *pvpfsd = opdata.p.fileinfo.vpfsd;
  *psffsi = opdata.p.fileinfo.sffsi;
  *psffsd = opdata.p.fileinfo.sffsd;

  end_request(opdata.request_id);
  return opdata.rc;
}


/******************************************************************************
**
** FS_FILEIO - Atomic I/O operations
**
** Parameters
** ----------
** struct sffsi far *psffsi       pointer to FSD independant file instance
** struct sffsd far *psffsd       pointer to FSD dependant file instance
** char far *pCmdList             UNVERIFIED pointer to information area
** unsigned short cbCmdList       size of area pointed to by pCmdList
** unsigned short far *poError    UNVERIFIED pointer to offset within
**                                pCmdList of command that caused an error
** unsigned short IOflag          bitfield of I/O suggestions
**   values:
**     IOFL_WRITETHRU   write all updated data before returning
**     IOFL_NOCACHE     don't cache any new data
**
******************************************************************************/

#pragma argsused
short int far pascal _export _loadds 
FS_FILEIO(struct sffsi far *psffsi, 
	  struct sffsd far *psffsd,
	  char far *pCmdList, 
	  unsigned short cbCmdList,
	  unsigned short far *poError, 
	  unsigned short IOflag)
{
  return ERROR_NOT_SUPPORTED;
}

/******************************************************************************
**
** FS_MOVE - Move/rename a file
**
** Parameters
** ----------
** struct cdfsi far *pcdfsi       pointer to FSD independant current directory
** struct cdfsd far *pcdfsd       pointer to FSD dependant current directory
** char far *pSrc                 pointer to source filename
** unsigned short iSrcCurDirEnd   offset to the end of the current directory 
**                                in pSrc
** char far *pDst                 pointer to destination filename
** unsigned short iDstCurDirEnd   offset to the end of the current directory 
**                                in pDst
** unsigned short flags           0x40 = destination is non 8.3 filename
**
******************************************************************************/

#pragma argsused
short int far pascal _export _loadds 
FS_MOVE(struct cdfsi far *pcdfsi, 
	struct cdfsd far *pcdfsd,
	char far *pSrc, 
	unsigned short iSrcCurDirEnd,
	char far *pDst, 
	unsigned short iDstCurDirEnd,
	unsigned short flags)
{
  unsigned short rc;
  OPDATA opdata;
  struct vpfsi *pvpfsi;
  struct vpfsd *pvpfsd;

  if (!CPAttached)
    return ERROR_NOT_READY;

  opdata.funccode = CPFC_MOVE;
  FSH_GETVOLPARM(pcdfsi->cdi_hVPB, &pvpfsi, &pvpfsd);
  opdata.p.move.vpfsd = *pvpfsd;
  opdata.p.move.cdfsi = *pcdfsi;
  opdata.p.move.cdfsd = *pcdfsd;
  strncpy(opdata.p.move.Src, pSrc, sizeof(opdata.p.move.Src));
  opdata.p.move.iSrcCurDirEnd = iSrcCurDirEnd;
  strncpy(opdata.p.move.Dst, pDst, sizeof(opdata.p.move.Dst));
  opdata.p.move.iDstCurDirEnd = iDstCurDirEnd;
  opdata.p.move.flags = flags;

  rc = do_request(&opdata, NULL, 0);
  if (rc != NO_ERROR)
    return rc;

  *pvpfsd = opdata.p.move.vpfsd;
  *pcdfsd = opdata.p.move.cdfsd;

  end_request(opdata.request_id);
  return opdata.rc;
}


/******************************************************************************
**
** FS_NEWSIZE - Change size of file
**
** Parameters
** ----------
** struct sffsi far *psffsi      pointer to FSD independant file instance
** struct sffsd far *psffsd      pointer to FSD dependant file instance
** unsigned long len             new length of file
** unsigned short IOflag         bitfield of I/O suggestions
**   values:
**     IOFL_WRITETHRU   write all updated data before returning
**     IOFL_NOCACHE     don't cache any new data
**
******************************************************************************/

#pragma argsused
short int far pascal _export _loadds 
FS_NEWSIZE(struct sffsi far *psffsi, 
	   struct sffsd far *psffsd,
	   unsigned long len, 
	   unsigned short IOflag)
{
  unsigned short rc;
  OPDATA opdata;
  struct vpfsi *pvpfsi;
  struct vpfsd *pvpfsd;

  if (!CPAttached)
    return ERROR_NOT_READY;

  opdata.funccode = CPFC_NEWSIZE;
  FSH_GETVOLPARM(psffsi->sfi_hVPB, &pvpfsi, &pvpfsd);
  opdata.p.newsize.vpfsd = *pvpfsd;
  opdata.p.newsize.sffsi = *psffsi;
  opdata.p.newsize.sffsd = *psffsd;
  opdata.p.newsize.len = len;
  opdata.p.newsize.IOflag = IOflag;

  rc = do_request(&opdata, NULL, 0);
  if (rc != NO_ERROR)
    return rc;

  *pvpfsd = opdata.p.newsize.vpfsd;
  *psffsi = opdata.p.newsize.sffsi;
  *psffsd = opdata.p.newsize.sffsd;

  end_request(opdata.request_id);
  return opdata.rc;
}


/******************************************************************************
**
** FS_OPENCREATE - Open or create a new file
**
** Parameters
** ----------
** struct cdfsi far *pcdfsi       pointer to FSD independant current directory
** struct cdfsd far *pcdfsd       pointer to FSD dependant current directory
** char far *pName                pointer to filename
** unsigned short iCurDirEnd      offset to the end of the current directory 
**                                in pName
** struct sffsi far *psffsi       pointer to FSD independant file instance
** struct sffsd far *psffsd       pointer to FSD dependant file instance
** unsigned long ulOpenMode       sharing and access mode
** unsigned short usOpenFlag      action to take when file exists/doesn't exist
** unsigned short far *pusAction  returns the action that the IFS took
** unsigned short usAttr          OS/2 file attributes
** char far *pEABuf               UNVERIFIED pointer to EAs to attach 
**                                to new file
** unsigned short far *pfgenflag  flags returned by the IFS
**   values:
**     FOC_NEEDEAS      indicates there are critical EAs associated 
**                      with the file
**
******************************************************************************/

#pragma argsused
short int far pascal _export _loadds 
FS_OPENCREATE(struct cdfsi far *pcdfsi, 
	      struct cdfsd far *pcdfsd,
	      char far *pName, 
	      unsigned short iCurDirEnd,
	      struct sffsi far *psffsi, 
	      struct sffsd far *psffsd,
	      unsigned long ulOpenMode, 
	      unsigned short usOpenFlag,
	      unsigned short far *pusAction, 
	      unsigned short usAttr,
	      char far *pEABuf, 
	      unsigned short far *pfgenflag)
{
  unsigned short rc;
  OPDATA opdata;
  struct vpfsi *pvpfsi;
  struct vpfsd *pvpfsd;
  char *pData;
  unsigned short data_len;

  /* Seems that pEABuf can be something like 0003:0000 although NULL pointer
     was passed to DosOpen. If offset is zero, we treat it as NULL. */
  if (FP_OFF(pEABuf)) {
    rc = FSH_PROBEBUF(PB_OPREAD, pEABuf, sizeof(EAOP));
    if (rc != NO_ERROR)
      return rc;
    else {
      /* Check that at least the first ULONG is accessible */
      rc = FSH_PROBEBUF(PB_OPREAD, ((EAOP *)pEABuf)->fpFEAList, sizeof(ULONG));
      if (rc != NO_ERROR)
        return rc;

      /* Check that it is <=64K */
      if (((EAOP *)pEABuf)->fpFEAList->cbList > 0xffff)
        return ERROR_INVALID_PARAMETER;

      /* Check that we can access the whole list */
      rc = FSH_PROBEBUF(PB_OPREAD, ((EAOP *)pEABuf)->fpFEAList,
                        ((EAOP *)pEABuf)->fpFEAList->cbList);
      if (rc != NO_ERROR)
        return rc;
    }
  }

  if (!CPAttached)
    return ERROR_NOT_READY;

  opdata.funccode = CPFC_OPENCREATE;
  FSH_GETVOLPARM(psffsi->sfi_hVPB, &pvpfsi, &pvpfsd);
  opdata.p.opencreate.vpfsd = *pvpfsd;
  opdata.p.opencreate.cdfsi = *pcdfsi;
  opdata.p.opencreate.cdfsd = *pcdfsd;
  strncpy(opdata.p.opencreate.Name, pName, 
	  sizeof(opdata.p.opencreate.Name));
  opdata.p.opencreate.iCurDirEnd = iCurDirEnd;
  opdata.p.opencreate.sffsi = *psffsi;
  opdata.p.opencreate.sffsd = *psffsd;
  opdata.p.opencreate.ulOpenMode = ulOpenMode;
  opdata.p.opencreate.usOpenFlag = usOpenFlag;
  opdata.p.opencreate.usAttr = usAttr;
  if (FP_OFF(pEABuf)) {
    pData = (char *)(((EAOP *)pEABuf)->fpFEAList);
    data_len = ((EAOP *)pEABuf)->fpFEAList->cbList;
  }
  else {
    pData = NULL;
    data_len = 0;
  }

  rc = do_request(&opdata, &pData, data_len);
  if (rc != NO_ERROR)
    return rc;

  *pvpfsd = opdata.p.opencreate.vpfsd;
  *pcdfsd = opdata.p.opencreate.cdfsd;
  *psffsi = opdata.p.opencreate.sffsi;
  *psffsd = opdata.p.opencreate.sffsd;
  *pusAction = opdata.p.opencreate.usAction;
  *pfgenflag = opdata.p.opencreate.fgenflag;
  if (FP_OFF(pEABuf))
    ((EAOP *)pEABuf)->oError = opdata.p.opencreate.oError;

  end_request(opdata.request_id);
  return opdata.rc;
}


/******************************************************************************
**
** FS_PATHINFO - get/set file information by filename
**
** Parameters
** ----------
** unsigned short flag           indicates flavor of call
**   values:
**     PI_RETRIEVE      retrieve information
**     PI_SET           set information
** struct cdfsi far *pcdfsi      pointer to FSD independant current directory
** struct cdfsd far *pcdfsd      pointer to FSD dependant current directory
** char far *pName               pointer to filename
** unsigned short iCurDirEnd     offset to the end of the current directory 
**                               in pName
** unsigned short level          level of information to get/set
** char far *pData               UNVERIFIED pointer to information area
** unsigned short cbData         size of area pointed to by pData
**
******************************************************************************/

#pragma argsused
short int far pascal _export _loadds 
FS_PATHINFO(unsigned short flag, 
	    struct cdfsi far *pcdfsi,
	    struct cdfsd far *pcdfsd, 
	    char far *pName,
	    unsigned short iCurDirEnd, 
	    unsigned short level,
	    char far *pData, 
	    unsigned short cbData)
{
  unsigned short rc;
  OPDATA opdata;
  struct vpfsi *pvpfsi;
  struct vpfsd *pvpfsd;
  char *pdata;
  unsigned short data_len;

  flag &= 0x01;    /* Ignore write-through flag */

  if (flag == PI_RETRIEVE && cbData != 0)
    rc = FSH_PROBEBUF(PB_OPWRITE, pData, cbData);
  else if (cbData != 0)
    rc = FSH_PROBEBUF(PB_OPREAD, pData, cbData);
  if (rc != NO_ERROR)
    return rc;

  /* Special EA check */
  if (flag == PI_RETRIEVE && (level == 3 || level == 4)) {
    /* Check to make sure the buffer's at least big enough for an EAOP */
    if (cbData < sizeof(EAOP))
      return ERROR_INVALID_PARAMETER;

    if(level==3) {
      /* Check accessability of GEA */
      
      rc = FSH_PROBEBUF(PB_OPREAD, ((EAOP *)pData)->fpGEAList, sizeof(ULONG));
      if (rc != NO_ERROR)
	return rc;
      
      if (((EAOP *)pData)->fpGEAList->cbList > 0xffff)
	return ERROR_INVALID_PARAMETER;

      if (((EAOP *)pData)->fpGEAList->cbList == 0)
	return ERROR_BAD_FORMAT;

      rc = FSH_PROBEBUF(PB_OPREAD, ((EAOP *)pData)->fpGEAList,
			((EAOP *)pData)->fpGEAList->cbList);
      if (rc != NO_ERROR)
	return rc;
    }

    /* Same thing for FEA */

    rc = FSH_PROBEBUF(PB_OPWRITE, ((EAOP *)pData)->fpFEAList, sizeof(ULONG));
    if (rc != NO_ERROR)
      return rc;

    if (((EAOP *)pData)->fpFEAList->cbList > 0xffff)
      return ERROR_INVALID_PARAMETER;

    rc = FSH_PROBEBUF(PB_OPWRITE, ((EAOP *)pData)->fpFEAList,
                      ((EAOP *)pData)->fpFEAList->cbList);
    if (rc != NO_ERROR)
      return rc;
  }

  if (flag == PI_SET && level == 2) {
    rc = FSH_PROBEBUF(PB_OPREAD, ((EAOP *)pData)->fpFEAList, sizeof(ULONG));
    if (rc != NO_ERROR)
      return rc;

    if (((EAOP *)pData)->fpFEAList->cbList > 0xffff)
      return ERROR_INVALID_PARAMETER;

    rc = FSH_PROBEBUF(PB_OPREAD, ((EAOP *)pData)->fpFEAList,
                      ((EAOP *)pData)->fpFEAList->cbList);
    if (rc != NO_ERROR)
      return rc;
  }

  if (!CPAttached)
    return ERROR_NOT_READY;

  opdata.funccode = CPFC_PATHINFO;
  FSH_GETVOLPARM(pcdfsi->cdi_hVPB, &pvpfsi, &pvpfsd);
  opdata.p.pathinfo.vpfsd = *pvpfsd;
  opdata.p.pathinfo.flag = flag;
  opdata.p.pathinfo.cdfsi = *pcdfsi;
  opdata.p.pathinfo.cdfsd = *pcdfsd;
  strncpy(opdata.p.pathinfo.Name, pName, 
	  sizeof(opdata.p.pathinfo.Name));
  opdata.p.pathinfo.iCurDirEnd = iCurDirEnd;
  opdata.p.pathinfo.level = level;

  /* Copy either the buffer, FEA or GEA list as appropriate */
  if (flag == PI_RETRIEVE && level == 3) {
    pdata = (char *)(((EAOP *)pData)->fpGEAList);
    data_len = ((EAOP *)pData)->fpGEAList->cbList;
    /* cbData is FEA list size, as FEA list will be returned */
    opdata.p.pathinfo.cbData = ((EAOP *)pData)->fpFEAList->cbList;
  }
  else if(level==4) {
    opdata.p.pathinfo.cbData = ((EAOP *)pData)->fpFEAList->cbList;
    pdata = NULL;
    data_len = 0;
  }
  else if (flag == FI_SET && level == 2) {
    pdata = (char *)(((EAOP *)pData)->fpFEAList);
    data_len = ((EAOP *)pData)->fpFEAList->cbList;
    opdata.p.pathinfo.cbData = ((EAOP *)pData)->fpFEAList->cbList;
  }
  else {
    pdata = pData;
    data_len = cbData;
    opdata.p.pathinfo.cbData = cbData;
  }

  rc = do_request(&opdata, &pdata, data_len);
  if (rc != NO_ERROR)
    return rc;

  if (flag == PI_RETRIEVE) {
    /* If level 3 or 4, copy results starting at 
       the oError entry in the EAOP */
    if (level != 3 && level != 4)
      memcpy(pData, pdata, cbData);
    else
      memcpy(((EAOP *)pData)->fpFEAList, pdata,
             ((FEALIST *)pData)->cbList);
  }
  if (flag == PI_RETRIEVE && level == 3 ||
      flag == PI_RETRIEVE && level == 4 ||
      flag == PI_SET && level == 2)
    ((EAOP *)pData)->oError = opdata.p.pathinfo.oError;
  *pvpfsd = opdata.p.pathinfo.vpfsd;
  *pcdfsd = opdata.p.pathinfo.cdfsd;

  end_request(opdata.request_id);
  return opdata.rc;
}


/******************************************************************************
**
** FS_READ - read data from a file
**
** Parameters
** ----------
** struct sffsi far *psffsi      pointer to FSD independant file instance
** struct sffsd far *psffsd      pointer to FSD dependant file instance
** char far *pData               UNVERIFIED pointer to buffer
** unsigned short far *pLen      length of buffer
** unsigned short IOflag         bitfield of I/O suggestions
**   values:
**     IOFL_WRITETHRU   write all updated data before returning
**     IOFL_NOCACHE     don't cache any new data
**
******************************************************************************/

#pragma argsused
short int far pascal _export _loadds 
FS_READ(struct sffsi far *psffsi, 
	struct sffsd far *psffsd,
	char far *pData, 
	unsigned short far *pLen,
	unsigned short IOflag)
{
  unsigned short rc;
  OPDATA opdata;
  struct vpfsi *pvpfsi;
  struct vpfsd *pvpfsd;
  char *pdata;

  if (*pLen != 0)
    rc = FSH_PROBEBUF(PB_OPWRITE, pData, *pLen);
  if (rc != NO_ERROR)
    return rc;

  if (!CPAttached)
    return ERROR_NOT_READY;

  opdata.funccode = CPFC_READ;
  FSH_GETVOLPARM(psffsi->sfi_hVPB, &pvpfsi, &pvpfsd);
  opdata.p.read.vpfsd = *pvpfsd;
  opdata.p.read.sffsi = *psffsi;
  opdata.p.read.sffsd = *psffsd;
  opdata.p.read.Len = *pLen;
  opdata.p.read.IOflag = IOflag;
  pdata = NULL;

  rc = do_request(&opdata, &pdata, 0);
  if (rc != NO_ERROR)
    return rc;

  /* Copy results */
  *pvpfsd = opdata.p.read.vpfsd;
  *psffsi = opdata.p.read.sffsi;
  *psffsd = opdata.p.read.sffsd;
  *pLen = opdata.p.read.Len;
  memcpy(pData, pdata, *pLen);

  end_request(opdata.request_id);
  return opdata.rc;
}


/******************************************************************************
**
** FS_WRITE - write data to a file
**
** Parameters
** ----------
** struct sffsi far *psffsi      pointer to FSD independant file instance
** struct sffsd far *psffsd      pointer to FSD dependant file instance
** char far *pData               UNVERIFIED pointer to buffer
** unsigned short far *pLen      length of buffer
** unsigned short IOflag         bitfield of I/O suggestions
**   values:
**     IOFL_WRITETHRU   write all updated data before returning
**     IOFL_NOCACHE     don't cache any new data
**
******************************************************************************/

#pragma argsused
short int far pascal _export _loadds 
FS_WRITE(struct sffsi far *psffsi, 
	 struct sffsd far *psffsd,
	 char far *pData, 
	 unsigned short far *pLen,
	 unsigned short IOflag)
{
  unsigned short rc;
  OPDATA opdata;
  struct vpfsi *pvpfsi;
  struct vpfsd *pvpfsd;
  char *pdata;

  if (*pLen != 0)
    rc = FSH_PROBEBUF(PB_OPREAD, pData, *pLen);
  if (rc != NO_ERROR)
    return rc;

  if (!CPAttached)
    return ERROR_NOT_READY;

  opdata.funccode = CPFC_WRITE;
  FSH_GETVOLPARM(psffsi->sfi_hVPB, &pvpfsi, &pvpfsd);
  opdata.p.write.vpfsd = *pvpfsd;
  opdata.p.write.sffsi = *psffsi;
  opdata.p.write.sffsd = *psffsd;
  opdata.p.write.Len = *pLen;
  opdata.p.write.IOflag = IOflag;
  pdata = pData;

  rc = do_request(&opdata, &pdata, *pLen);
  if (rc != NO_ERROR)
    return rc;

  *pvpfsd = opdata.p.write.vpfsd;
  *psffsi = opdata.p.write.sffsi;
  *psffsd = opdata.p.write.sffsd;
  *pLen = opdata.p.write.Len;

  end_request(opdata.request_id);
  return opdata.rc;
}



/*-----------------------------------------------------------------------------
--
-- Directory management
--
-----------------------------------------------------------------------------*/



/******************************************************************************
**
** FS_FINDCLOSE - End a directory search
**
** Parameters
** ----------
** struct fsfsi far *pfsfsi      pointer to FSD independant search record
** struct fsfsd far *pfsfsd      pointer to FSD dependant search record
**
******************************************************************************/

#pragma argsused
short int far pascal _export _loadds 
FS_FINDCLOSE(struct fsfsi far *pfsfsi, 
	     struct fsfsd far *pfsfsd)
{
  unsigned short rc;
  OPDATA opdata;
  struct vpfsi *pvpfsi;
  struct vpfsd *pvpfsd;

  if (!CPAttached)
    return ERROR_NOT_READY;

  opdata.funccode = CPFC_FINDCLOSE;
  FSH_GETVOLPARM(pfsfsi->fsi_hVPB, &pvpfsi, &pvpfsd);
  opdata.p.findclose.vpfsd = *pvpfsd;
  opdata.p.findclose.fsfsi = *pfsfsi;
  opdata.p.findclose.fsfsd = *pfsfsd;

  rc = do_request(&opdata, NULL, 0);
  if (rc != NO_ERROR)
    return rc;

  *pfsfsd = opdata.p.findclose.fsfsd;

  end_request(opdata.request_id);
  return opdata.rc;
}


/******************************************************************************
**
** FS_FINDFIRST - Begin a new directory search
**
** Parameters
** ----------
** struct cdfsi far *pcdfsi      pointer to FSD independant current directory
** struct cdfsd far *pcdfsd      pointer to FSD dependant current directory
** char far *pName               pointer to filename mask
** unsigned short iCurDirEnd     offset to the end of the current directory 
**                               in pName
** unsigned short attr           attribute mask
** struct fsfsi far *pfsfsi      pointer to FSD independant search record
** struct fsfsd far *pfsfsd      pointer to FSD dependant search record
** char far *pData               UNVERIFIED pointer to information area
** unsigned short cbData         size of area pointed to by pData
** unsigned short far *pcMatch   maximum number of entries to return*
**                               number of entries actually returned
** unsigned short level          level of information to return
** unsigned short flags          indicates whether to return 
**                               position information
**   values:
**     FF_NOPOS         don't return any position information
**     FF_GETPOS        return position information in buffer
**
******************************************************************************/

#pragma argsused
short int far pascal _export _loadds 
FS_FINDFIRST(struct cdfsi far *pcdfsi, 
	     struct cdfsd far *pcdfsd,
	     char far *pName, 
	     unsigned short iCurDirEnd,
	     unsigned short attr, 
	     struct fsfsi far *pfsfsi,
	     struct fsfsd far *pfsfsd, 
	     char far *pData,
	     unsigned short cbData, 
	     unsigned short far *pcMatch,
	     unsigned short level,
	     unsigned short flags)
{
  unsigned short rc;
  OPDATA opdata;
  struct vpfsi *pvpfsi;
  struct vpfsd *pvpfsd;
  char *pdata;
  unsigned short data_len;

  /* Verify parameters */
  if (cbData != 0)
    rc = FSH_PROBEBUF(PB_OPWRITE, pData, cbData);
  if (rc != NO_ERROR)
    return rc;

  /* Special EA check */
  if (level == 3 || level == 4) {
    /* Check to make sure the buffer's at least big enough for an EAOP */
    if (cbData < sizeof(EAOP))
      return ERROR_INVALID_PARAMETER;

    if(level==3) {
      /* Check that at least the first ULONG is accessible */
      rc = FSH_PROBEBUF(PB_OPREAD, ((EAOP *)pData)->fpGEAList, sizeof(ULONG));
      if (rc != NO_ERROR)
	return rc;
      
      /* Check that it is <=64K */
      if (((EAOP *)pData)->fpGEAList->cbList > 0xffff)
	return ERROR_INVALID_PARAMETER;

      if (((EAOP *)pData)->fpGEAList->cbList == 0)
	return ERROR_BAD_FORMAT;

      /* Check that we can access the whole list */
      rc = FSH_PROBEBUF(PB_OPREAD, ((EAOP *)pData)->fpGEAList,
			((EAOP *)pData)->fpGEAList->cbList);
      if (rc != NO_ERROR)
	return rc;
    }
  }

  if (!CPAttached)
    return ERROR_NOT_READY;

  /* Setup parameters */
  opdata.funccode = CPFC_FINDFIRST;
  FSH_GETVOLPARM(pfsfsi->fsi_hVPB, &pvpfsi, &pvpfsd);
  opdata.p.findfirst.vpfsd = *pvpfsd;
  opdata.p.findfirst.cdfsi = *pcdfsi;
  opdata.p.findfirst.cdfsd = *pcdfsd;
  strncpy(opdata.p.findfirst.Name, pName, 
	  sizeof(opdata.p.findfirst.Name));
  opdata.p.findfirst.iCurDirEnd = iCurDirEnd;
  opdata.p.findfirst.attr = attr;
  opdata.p.findfirst.fsfsi = *pfsfsi;
  opdata.p.findfirst.fsfsd = *pfsfsd;
  /* Copy either the buffer or the GEA list as appropriate */
  if (level==3) {
    pdata = (char *)(((EAOP *)pData)->fpGEAList);
    data_len = ((EAOP *)pData)->fpGEAList->cbList;
    opdata.p.findfirst.cbData = cbData-sizeof(EAOP);
  }
  else if(level==4) {
    pdata = NULL;
    data_len = 0;
    opdata.p.findfirst.cbData = cbData-sizeof(EAOP);
  }
  else {
    pdata = pData;
    data_len = cbData;
    opdata.p.findfirst.cbData = cbData;
  }
  opdata.p.findfirst.cMatch = *pcMatch;
  opdata.p.findfirst.level = level;
  opdata.p.findfirst.flags = flags;

  rc = do_request(&opdata, &pdata, data_len);
  if (rc != NO_ERROR)
    return rc;

  /* Copy results */
  *pcdfsd = opdata.p.findfirst.cdfsd;
  *pfsfsd = opdata.p.findfirst.fsfsd;
  /* If level 3, copy results starting at the oError entry in the EAOP */
  if(level==3 || level==4) {
    memcpy(pData + sizeof(EAOP), pdata, cbData - sizeof(EAOP));
    ((EAOP *)pData)->oError = opdata.p.findfirst.oError;
  }
  else
    memcpy(pData, pdata, cbData);
  *pcMatch = opdata.p.findfirst.cMatch;
  
  end_request(opdata.request_id);
  return opdata.rc;
}


/******************************************************************************
**
** FS_FINDFROMNAME - Restart directory search
**
** Parameters
** ----------
** struct fsfsi far *pfsfsi      pointer to FSD independant search record
** struct fsfsd far *pfsfsd      pointer to FSD dependant search record
** char far *pData               UNVERIFIED pointer to information area
** unsigned short cbData         size of area pointed to by pData
** unsigned short far *pcMatch   maximum number of entries to return/
**                               number of entries actually returned
** unsigned short level          level of information to return
** unsigned long position        position in directory to restart search from
** char far *pName               pointer to filename to restart search from
** unsigned short flags          indicates whether to return 
**                               position information
**   values:
**     FF_NOPOS         don't return any position information
**     FF_GETPOS        return position information in buffer
**
******************************************************************************/

#pragma argsused
short int far pascal _export _loadds 
FS_FINDFROMNAME(struct fsfsi far *pfsfsi,
		struct fsfsd far *pfsfsd,
		char far *pData, 
		unsigned short cbData,
		unsigned short far *pcMatch, 
		unsigned short level,
		unsigned long position, 
		char far *pName,
		unsigned short flags)
{
  unsigned short rc;
  OPDATA opdata;
  struct vpfsi *pvpfsi;
  struct vpfsd *pvpfsd;
  char *pdata;
  unsigned short data_len;
  
  if (cbData != 0)
    rc = FSH_PROBEBUF(PB_OPWRITE, pData, cbData);
  if (rc != NO_ERROR)
    return rc;
  
  if (level == 3 || level == 4) {
    if (cbData < sizeof(EAOP))
      return ERROR_INVALID_PARAMETER;
    
    if(level==3) {
      rc = FSH_PROBEBUF(PB_OPREAD, ((EAOP *)pData)->fpGEAList, sizeof(ULONG));
      if (rc != NO_ERROR)
	return rc;
      
      if (((EAOP *)pData)->fpGEAList->cbList > 0xffff)
	return ERROR_INVALID_PARAMETER;
      
      if (((EAOP *)pData)->fpGEAList->cbList == 0)
	return ERROR_BAD_FORMAT;

      rc = FSH_PROBEBUF(PB_OPREAD, ((EAOP *)pData)->fpGEAList, 
			((EAOP *)pData)->fpGEAList->cbList);
      if (rc != NO_ERROR)
	return rc;
    }
  }
  
  if (!CPAttached)
    return ERROR_NOT_READY;

  opdata.funccode = CPFC_FINDFROMNAME;
  FSH_GETVOLPARM(pfsfsi->fsi_hVPB, &pvpfsi, &pvpfsd);
  opdata.p.findfromname.vpfsd = *pvpfsd;
  opdata.p.findfromname.fsfsi = *pfsfsi;
  opdata.p.findfromname.fsfsd = *pfsfsd;
  if (level==3) {
    pdata = (char *)(((EAOP *)pData)->fpGEAList);
    data_len = ((EAOP *)pData)->fpGEAList->cbList;
    opdata.p.findfromname.cbData = cbData-sizeof(EAOP);
  }
  else if(level==4) {
    pdata = NULL;
    data_len = 0;
    opdata.p.findfromname.cbData = cbData-sizeof(EAOP);
  }
  else {
    pData = pdata;
    data_len = cbData;
    opdata.p.findfromname.cbData = cbData;
  }
  opdata.p.findfromname.cMatch = *pcMatch;
  opdata.p.findfromname.level = level;
  opdata.p.findfromname.position = position;
  strncpy(opdata.p.findfromname.Name, pName, 
	  sizeof(opdata.p.findfromname.Name));
  opdata.p.findfromname.flags = flags;
  
  rc = do_request(&opdata, &pdata, data_len);
  if (rc != NO_ERROR)
    return rc;
  
  *pfsfsd = opdata.p.findfromname.fsfsd;
  if(level==3 || level==4) {
    memcpy(pData + sizeof(EAOP), pdata, cbData - sizeof(EAOP));
    ((EAOP *)pData)->oError = opdata.p.findfromname.oError;
  }
  else
    memcpy(pData, pdata, cbData);
  *pcMatch = opdata.p.findfromname.cMatch;
  
  end_request(opdata.request_id);
  return opdata.rc;
}


/******************************************************************************
**
** FS_FINDNEXT - Continue directory search
**
** Parameters
** ----------
** struct fsfsi far *pfsfsi      pointer to FSD independant search record
** struct fsfsd far *pfsfsd      pointer to FSD dependant search record
** char far *pData               UNVERIFIED pointer to information area
** unsigned short cbData         size of area pointed to by pData
** unsigned short far *pcMatch   maximum number of entries to return*
**                               number of entries actually returned
** unsigned short level          level of information to return
** unsigned short flags          indicates whether to return 
**                               position information
**   values:
**     FF_NOPOS         don't return any position information
**     FF_GETPOS        return position information in buffer
**
******************************************************************************/

#pragma argsused
short int far pascal _export _loadds 
FS_FINDNEXT(struct fsfsi far *pfsfsi, 
	    struct fsfsd far *pfsfsd,
	    char far *pData, 
	    unsigned short cbData,
	    unsigned short far *pcMatch, 
	    unsigned short level,
	    unsigned short flags)
{
  unsigned short rc;
  OPDATA opdata;
  struct vpfsi *pvpfsi;
  struct vpfsd *pvpfsd;
  char *pdata;
  unsigned short data_len;
   
  if (cbData != 0)
    rc = FSH_PROBEBUF(PB_OPWRITE, pData, cbData);
  if (rc != NO_ERROR)
    return rc;

  if (level == 3 || level == 4) {
    if (cbData < sizeof(EAOP))
      return ERROR_INVALID_PARAMETER;

    if(level==3) {
      rc = FSH_PROBEBUF(PB_OPREAD, ((EAOP *)pData)->fpGEAList, sizeof(ULONG));
      if (rc != NO_ERROR)
	return rc;
      
      if (((EAOP *)pData)->fpGEAList->cbList > 0xffff)
	return ERROR_INVALID_PARAMETER;
      
      if (((EAOP *)pData)->fpGEAList->cbList == 0)
	return ERROR_BAD_FORMAT;

      rc = FSH_PROBEBUF(PB_OPREAD, ((EAOP *)pData)->fpGEAList, 
			((EAOP *)pData)->fpGEAList->cbList);
      if (rc != NO_ERROR)
	return rc;
    }
  }

  if (!CPAttached)
    return ERROR_NOT_READY;

  opdata.funccode = CPFC_FINDNEXT;
  FSH_GETVOLPARM(pfsfsi->fsi_hVPB, &pvpfsi, &pvpfsd);
  opdata.p.findnext.vpfsd = *pvpfsd;
  opdata.p.findnext.fsfsi = *pfsfsi;
  opdata.p.findnext.fsfsd = *pfsfsd;
  if (level==3) {
    pdata = (char *)(((EAOP *)pData)->fpGEAList);
    data_len = ((EAOP *)pData)->fpGEAList->cbList;
    opdata.p.findnext.cbData = cbData-sizeof(EAOP);
  }
  else if(level==4) {
    pdata = NULL;
    data_len = 0;
    opdata.p.findnext.cbData = cbData-sizeof(EAOP);
  }
  else {
    pdata = pData;
    data_len = cbData;
    opdata.p.findnext.cbData = cbData;
  }
  opdata.p.findnext.cMatch = *pcMatch;
  opdata.p.findnext.level = level;
  opdata.p.findnext.flags = flags;

  rc = do_request(&opdata, &pdata, data_len);
  if (rc != NO_ERROR)
    return rc;

  *pfsfsd = opdata.p.findnext.fsfsd;
  if(level==3 || level==4) {
    memcpy(pData + sizeof(EAOP), pdata, cbData - sizeof(EAOP));
    ((EAOP *)pData)->oError = opdata.p.findnext.oError;
  }
  else
    memcpy(pData, pdata, cbData);
  *pcMatch = opdata.p.findnext.cMatch;

  end_request(opdata.request_id);
  return opdata.rc;
}


/******************************************************************************
**
** FS_FINDNOTIFYCLOSE - End a directory update request
**
** Parameters
** ----------
** unsigned short handle                handle of update request to close
**
******************************************************************************/

#pragma argsused
short int far pascal _export _loadds 
FS_FINDNOTIFYCLOSE(unsigned short handle)
{
  return ERROR_NOT_SUPPORTED;
}


/******************************************************************************
**
** FS_FINDNOTIFYFIRST - Begin a new directory update request
**
** Parameters
** ----------
** struct cdfsi far *pcdfsi             pointer to FSD independant current directory
** struct cdfsd far *pcdfsd             pointer to FSD dependant current directory
** char far *pName                      pointer to filename mask
** unsigned short iCurDirEnd            offset to the end of the current directory in pName
** unsigned short attr                  attribute mask
** unsigned short far *pHandle          pointer to place where FSD stores its handle
** char far *pData                      UNVERIFIED pointer to information area
** unsigned short cbData                size of area pointed to by pData
** unsigned short far *pcMatch          maximum number of entries to return*
**                                      number of entries actually returned
** unsigned short level                 level of information to return
** unsigned long timeout                timeout in milliseconds
**
******************************************************************************/

#pragma argsused
short int far pascal _export _loadds 
FS_FINDNOTIFYFIRST(struct cdfsi far *pcdfsi, 
		   struct cdfsd far *pcdfsd,
		   char far *pName,
		   unsigned short iCurDirEnd,
		   unsigned short attr, 
		   unsigned short far *pHandle,
		   char far *pData, 
		   unsigned short cbData,
		   unsigned short far *pcMatch, 
		   unsigned short level,
		   unsigned long timeout)
{
  return ERROR_NOT_SUPPORTED;
}


/******************************************************************************
**
** FS_FINDNOTIFYNEXT - Continue directory update request
**
** Parameters
** ----------
** unsigned short handle                directory update handle
** char far *pData                      UNVERIFIED pointer to information area
** unsigned short cbData                size of area pointed to by pData
** unsigned short far *pcMatch          maximum number of entries to return*
**                                      number of entries actually returned
** unsigned short level                 level of information to return
** unsigned long timeout                timeout in milliseconds
**
******************************************************************************/

#pragma argsused
short int far pascal _export _loadds 
FS_FINDNOTIFYNEXT(unsigned short handle, 
		  char far *pData,
		  unsigned short cbData, 
		  unsigned short far *pcMatch,
		  unsigned short level, 
		  unsigned long timeout)
{
  return ERROR_NOT_SUPPORTED;
}



/*-----------------------------------------------------------------------------
--
-- FSD Extended Interface
--
-----------------------------------------------------------------------------*/



/******************************************************************************
**
** FS_IOCTL - Perform an IOCTL on the IFS
**
** Parameters
** ----------
** struct sffsi far *psffsi           pointer to FSD independant file instance
** struct sffsd far *psffsd           pointer to FSD dependant file instance
** unsigned short cat                 function category
** unsigned short func                function to perform
** char far *pParm                    UNVERIFIED pointer to parameter area
** unsigned short lenParm             size of area pointed to by pParm
** unsigned short far *pParmLenInOut  length of parameters passed in pParm
** char far *pData                    UNVERIFIED pointer to information area
** unsigned short lenData             size of area pointed to by pData
** unsigned short far *pDataLenInOut  length of parameters passed in pData
**
******************************************************************************/

#pragma argsused
short int far pascal _export _loadds 
FS_IOCTL(struct sffsi far *psffsi, 
	 struct sffsd far *psffsd,
	 unsigned short cat, 
	 unsigned short func,
	 char far *pParm, 
	 unsigned short lenParm,
	 unsigned far *pParmLenInOut, 
	 char far *pData,
	 unsigned short lenData, 
	 unsigned far *pDataLenInOut)
{
  struct vpfsi *pvpfsi;
  struct vpfsd *pvpfsd;
  int rc;

  FSH_GETVOLPARM(psffsi->sfi_hVPB, &pvpfsi, &pvpfsd);
  if(pParmLenInOut==NULL && pDataLenInOut==NULL)
    rc = FSH_DEVIOCTL(0, pvpfsi->vpi_hDEV, psffsi->sfi_selfsfn, cat, func, 
		      pParm, lenParm, pData, lenData);
  else {
    rc = FSH_DEVIOCTL(0, pvpfsi->vpi_hDEV, psffsi->sfi_selfsfn, cat, func, 
		      pParm, *pParmLenInOut, pData, *pDataLenInOut);
    if(pParmLenInOut)
      *pParmLenInOut = lenParm;
    if(pDataLenInOut)
      *pDataLenInOut = lenData;
  }
  return rc;
}

/*-----------------------------------------------------------------------------
--
-- Miscellaneous Functions
--
-----------------------------------------------------------------------------*/



/******************************************************************************
**
** FS_NMPIPE - Perform a named pipe operation
**
** Parameters
** ----------
** struct sffsi far *psffsi             pointer to FSD independant file instance
** struct sffsd far *psffsd             pointer to FSD dependant file instance
** unsigned short OpType                operation to perform
**   values:
**     NMP_GetPHandState
**     NMP_SetPHandState
**     NMP_PipeQInfo
**     NMP_PeekPipe
**     NMP_ConnectPipe
**     NMP_DisconnectPipe
**     NMP_TransactPipe
**     NMP_READRAW
**     NMP_WRITERAW
**     NMP_WAITPIPE
**     NMP_CALLPIPE
**     NMP_QNmPipeSemState
** union npoper far *pOpRec             data for operation
** char far *pData                      pointer to user data
** char far *pName                      pointer to remote named pipe
**
******************************************************************************/

#pragma argsused
short int far pascal _export _loadds 
FS_NMPIPE(struct sffsi far *psffsi, 
	  struct sffsd far *psffsd,
	  unsigned short OpType, 
	  union npoper far *pOpRec,
	  char far *pData, 
	  char far *pName)
{
  return ERROR_NOT_SUPPORTED;
}


/******************************************************************************
**
** FS_PROCESSNAME - Canonicalize a filename
**
** Parameters
** ----------
** char far *pNameBuf                   filename to canonicalize
**
******************************************************************************/

#pragma argsused
short int far pascal _export _loadds 
FS_PROCESSNAME(char far *pNameBuf)
{
  unsigned short rc;
  OPDATA opdata;

  if (!CPAttached)
    return ERROR_NOT_READY;

  opdata.funccode = CPFC_PROCESSNAME;
  strncpy(opdata.p.processname.NameBuf, pNameBuf, 
	  sizeof(opdata.p.processname.NameBuf));

  rc = do_request(&opdata, NULL, 0);
  if (rc != NO_ERROR)
    return rc;

  strcpy(pNameBuf, opdata.p.processname.NameBuf);

  end_request(opdata.request_id);
  return opdata.rc;
}


/******************************************************************************
**
** FS_SETSWAP - Inform IFS that it owns the swap file
**
** Parameters
** ----------
** struct sffsi far *psffsi       pointer to FSD independant file instance
** struct sffsd far *psffsd       pointer to FSD dependant file instance
**
******************************************************************************/

#pragma argsused
short int far pascal _export _loadds 
FS_SETSWAP(struct sffsi far *psffsi,
	   struct sffsd far *psffsd)
{
  return NO_ERROR;
}
