/*
 * HFS/2 - A Hierarchical File System Driver for OS/2
 * Copyright (C) 1996, 1997 Marcus Better
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#define INCL_DOSERRORS
#include <os2.h>

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>

#include "libhfs.h"
#include "os.h"
#include "ifs.h"
#include "../diskfunc.h"

/*
 * NAME:	os->open()
 * DESCRIPTION:	open a new descriptor from the given path and mode
 */
int os_open(void **priv, const char *path, int mode)
{
  PVOLINFO vol;
  unsigned short hVPB;
  unsigned short sector_size;

  if (sscanf(path, "hVPB=%hu,sector_size=%hu", &hVPB, &sector_size) != 2)
    ERROR(EACCES, 0);

  vol = malloc(sizeof(VOLINFO));
  vol->hVPB = hVPB;
  vol->sector_size = sector_size;
  vol->offset = 0;

  *priv = (void *)vol;
  return 0;

fail:
  return -1;
}

/*
 * NAME:	os->same()
 * DESCRIPTION:	return 1 iff path is same as the open descriptor
 */
int os_same(void **priv, const char *path)
{
  return 0;
}

/*
 * NAME:	os->lock()
 * DESCRIPTION:	lock a descriptor against concurrent access
 */
int os_lock(void **priv, int readonly)
{
  return 0;
}

/*
 * NAME:	os->close()
 * DESCRIPTION:	close an open descriptor
 */
int os_close(void **priv)
{
  free(*priv);
  *priv = NULL;
  return 0;
}

/*
 * NAME:	os->seek()
 * DESCRIPTION:	set a descriptor's seek pointer
 */
long os_seek(void **priv, long offset, int from)
{
  PVOLINFO vol = (PVOLINFO)*priv;

  if (from != HFS_SEEK_SET || offset%HFS_BLOCKSZ != 0)
    ERROR(EACCES, 0);

  vol->offset = offset;
  return offset;

fail:
  return -1;
}

/*
 * NAME:	os->read()
 * DESCRIPTION:	read bytes from an open descriptor
 */
long os_read(void **priv, void *buf, unsigned long len)
{
  PVOLINFO vol = (PVOLINFO)*priv;

  if (read_blocks(vol->offset/HFS_BLOCKSZ,
		  len/HFS_BLOCKSZ,
		  HFS_BLOCKSZ,
		  vol->sector_size,
		  vol->hVPB,
		  buf) != NO_ERROR)
    ERROR(EIO, 0);

  return len;

fail:
  return -1;
}

/*
 * NAME:	os->write()
 * DESCRIPTION:	write bytes to an open descriptor
 */
long os_write(void **priv, const void *buf, unsigned long len)
{
  PVOLINFO vol = (PVOLINFO)*priv;

  if (write_blocks(vol->offset/HFS_BLOCKSZ,
		   len/HFS_BLOCKSZ,
		   HFS_BLOCKSZ,
		   vol->sector_size,
		   vol->hVPB,
		   (void *)buf) != NO_ERROR)
    ERROR(EIO, 0);

  return len;

fail:
  return -1;
}
