#
# Rules.make
# This file contains rules which are shared between multiple Makefiles.
#

#
# False targets.
#
.PHONY: dummy sub_dirs clean dep

#
# Make subdirectories
#
all:: sub_dirs

#
# Common rules
#

%.s: %.c
	$(CC) $(CFLAGS) $(EXTRA_CFLAGS) -S $< -o $@

%.o: %.c
	$(CC) $(CFLAGS) $(EXTRA_CFLAGS) -c -o $@ $<

%.obj: %.c
	$(CC) $(CFLAGS) $(EXTRA_CFLAGS) -Zomf -c -o $@ $<

%.o: %.s
	$(AS) $(ASFLAGS) $(EXTRA_ASFLAGS) -o $@ $<

%.lib: %.a
	$(EMXOMF) $(EMXOMF_FLAGS) -o $@ $<

%.inf: %.ipf
	$(IPFC) /inf $<

%.o16: %.c
	$(C16) $(C16FLAGS) $(EXTRA_C16FLAGS) -c $< -o$@

#
# Rule to compile a set of .o files into one .o file
#
ifdef O_TARGET
all:: $(O_TARGET)

$(O_TARGET): $(O_OBJS)
	if exist $@ del $@
ifneq "$(strip $(O_OBJS))" ""
	$(LD) $(EXTRA_LDFLAGS) -r -o $@ $(O_OBJS)
else
	$(AR) rcs $@
endif
endif

#
# Rule to compile a set of .o files into one .a file
#
ifdef L_TARGET
all:: $(L_TARGET)

$(L_TARGET): $(L_OBJS)
	if exist $@ del $@
	$(AR) $(EXTRA_ARFLAGS) rcs $@ $(L_OBJS)
endif

#
# Make dependencies
#
dep:
	if exist *.c gcc -M -I$(CINCL_DEP) $(CFLAGS) $(EXTRA_CFLAGS) *.c > .depend
ifdef SUB_DIRS
	for %i in ($(SUB_DIRS)) do $(MAKE) -C %i dep
endif

#depend:
#	( sed -n '1,/^### DEPEND/p' Makefile;  \
#	  echo;  \
#	  $(CC) -MM $(CFLAGS) *.c;  \
#	) > Makefile.new
#	mv -f Makefile.new Makefile

#
# A rule to make subdirectories
#
sub_dirs: dummy
ifdef SUB_DIRS
	for %i in ($(SUB_DIRS)) do $(MAKE) -C %i
endif

#
# Clean up
#
clean::
ifdef SUB_DIRS
	for %i in ($(SUB_DIRS)) do $(MAKE) -C %i clean
endif

#
# A rule to do nothing
#
dummy:

#
# include dependency file
#
include .depend
